#ifndef _DATE_CPP
#define _DATE_CPP

#include <fstream>
#include <ctime>
#include "Date.h"

//===========================  ============================

Date::Date()
{
//   cout << "in Date default constructor" << endl;
   dYear = 1;  dMonth = 1;  dDay = 1;
                                 // 1    1 ..
   ++objCnt;
} // Date::Date

Date::Date( int yp, int mp, int dp )
{
//   cout << "in Date default constructor: ";
   if ( yp <= 0 )
      throw DateXptn( "Date", DateXptn::yearErr, yp );
   dYear = yp;
   if ( mp <= 0 || 12 < mp )
      throw DateXptn( "Date", DateXptn::monthErr, mp );
   dMonth = mp;
   if ( dp <= 0 || lastDay(dYear, dMonth) < dp )
      throw DateXptn( "Date", DateXptn::dayErr, yp, mp, dp );
   dDay = dp;
//   cout << int(dDay) << '.' << int(dMonth) << '.' << int(dYear) << endl;
   ++objCnt;
} // Date

Date::Date( const Date& d )
{
   dYear = d.dYear;  dMonth = d.dMonth;  dDay = d.dDay;
   ++objCnt;
} // Date::Date

//============================ getters ==============================
inline unsigned int Date::getYear() const {  return dYear;  }

inline unsigned int Date::getMonth() const {  return dMonth;  }

inline unsigned int Date::getDay() const {  return dDay;  }

//============================ setters ==============================
void Date::setYear( int ay )
{
   if ( ay <= 0 )
      throw DateXptn( "setYear", DateXptn::yearErr, ay ); 
   if ( dDay <= 0 || lastDay(ay, dMonth) < dDay )
      throw DateXptn( "setYear", DateXptn::dayErr,
                      ay, dMonth, dDay );
   dYear = ay;
} // Date::setYear

void Date::setMonth( int am )
{
   if ( am <= 0 || 12 < am )
      throw DateXptn( "setMonth", DateXptn::monthErr, am ); 
   if ( dDay <= 0 || lastDay(dYear, am) < dDay )
      throw DateXptn( "setMonth", DateXptn::dayErr,
                      dYear, am, dDay );
   dMonth = am;
} // Date::setMonth

void Date::setDay( int ad )
{
   if ( ad <= 0 || lastDay(dYear, dMonth) < ad )
      throw DateXptn( "setDay", DateXptn::dayErr,
                      dYear, dMonth, ad );
   dDay = ad;
} // Date::setDay

//=========================   ===========================
/*
Date& Date::operator++()
{
   if ( dDay < lastDay(dYear, dMonth) )
      ++dDay; 
   else //  
   {
      dDay = 1;
      if ( dMonth < 12 )
         ++dMonth; 
      else //  
      {
         dMonth = 1;
         ++dYear;
      } // if ( dMonth...
   } // if ( dDay...
   return *this;
} // Date::operator++
*/
Date& Date::forward( long int dd )
{
   const unsigned long secsPerDay = 86400;
   tm currD = { 0, 0, 12, dDay, dMonth-1, dYear-1900, 0, 0, 0 };
   time_t currT( mktime(&currD) );
   dd *= secsPerDay;     //   
   if ( dd < 0 )
   {
      if ( currT < (-dd) )  // currT + dd < 0
         throw DateXptn( "operator+=", DateXptn::outOfLimits,
                         dd/secsPerDay, *this );
   }
   else // dd >= 0
   {
      if ( dd > LONG_MAX - currT )  // currT + dd > LONG_MAX
         throw DateXptn( "forward", DateXptn::outOfLimits,
                         dd/secsPerDay, *this );
   }
   currT += dd;
   currD = *localtime( &currT );
   dYear = currD.tm_year+1900;
   dMonth = currD.tm_mon+1;
   dDay = currD.tm_mday;
   return *this;
} // Date::forward
/*
Date& Date::operator+=( long int dd )
{
   tm currD = { 0, 0, 12, dDay, dMonth-1, dYear-1900, 0, 0, 0 };
   time_t currT( mktime(&currD) );
   dd *= 86400;     //   
   if ( dd < 0 )
   {
      if ( currT < (-dd) )  // currT + dd < 0
         throw DateXptn( "operator+=", DateXptn::outOfLimits,
                         dd/86400, *this );
   }
   else // dd >= 0
   {
      if ( dd > LONG_MAX - currT )  // currT + dd > LONG_MAX
         throw DateXptn( "operator+=", DateXptn::outOfLimits,
                         dd/86400, *this );
   }
   currT += dd;
   currD = *localtime( &currT );
   dYear = currD.tm_year+1900;
   dMonth = currD.tm_mon+1;
   dDay = currD.tm_mday;
   return *this;
} // Date::operator+=
*/
void Date::load( istream& bin )
{
   Date tmp;
   bin.read( reinterpret_cast<char*>(&tmp.dYear), sizeof(dYear) );
   bin.read( reinterpret_cast<char*>(&tmp.dMonth), sizeof(dMonth) );
   bin.read( reinterpret_cast<char*>(&tmp.dDay), sizeof(dDay) ); 
   if ( bin.fail() )
      throw DateXptn( "load", DateXptn::cannotRead );
   *this = tmp;
}; // Date::load

void Date::save( ostream& bout ) const
{
   if ( bout.fail() )
      throw DateXptn( "save", DateXptn::fileNotOpen );
   bout.write( reinterpret_cast<const char*>(dYear),
               sizeof(dYear) );
   bout.write( reinterpret_cast<const char*>(dMonth),
               sizeof(dMonth) );
   bout.write( reinterpret_cast<const char*>(dDay),
               sizeof(dDay) ); 
   if ( bout.fail() )
      throw DateXptn( "save", DateXptn::cannotWrite );
}; // Date::save

//=====================   ======================
bool Date::isValidDate( int ay, int am, int ad )
{
   bool fv( true );
   try{  Date( ay, am, ad );  }
   catch( DateXptn& ) {  fv = false;  }
   return fv;
} // isValidDate

bool Date::isLeapYear( int y )
{
   bool fv( false );

   if ( y%400 == 0 )                   fv = true;
   else if ( y%4 == 0 && y%100 != 0 )  fv = true; 
   return fv;
} // isIsLeapYear
   
unsigned int Date::lastDay( int y, int m )
{
   unsigned int fv;

   if ( m == 1 || m == 3 || m == 5 || m == 7 || m == 8 ||
        m == 10 || m == 12 )
      fv = 31;
   else if ( m == 4 || m == 6 || m == 9 || m == 11 )
      fv = 30;
   else // m == 2
      if ( isLeapYear(y) )  fv = 29;
                      else  fv = 28;
   return fv;
} // lastDay

//======================   ======================

ostream& operator<<( ostream& tout, const Date& rhs )
{
   return tout << rhs.getDay() << '.' << rhs.getMonth() << '.'
               << rhs.getYear();
} // operator<<( ostream& tout, const Date

Date operator+( const Date& lhs, int rhs )
{
   Date fv( lhs );
   fv += rhs;
   return fv;
} // Date operator+

Date operator+( int lhs, const Date& rhs )
{
   return ( rhs + lhs );
} // Date operator+

bool operator==( const Date& lhs, const Date& rhs )
{
   return ( lhs.getYear() == rhs.getYear() &&
            lhs.getMonth() == rhs.getMonth() &&
            lhs.getDay() == rhs.getDay() );
}; // operator==( const Date

bool operator!=( const Date& lhs, const Date& rhs )
{  return !(lhs == rhs);  }

bool operator<( const Date& lhs, const Date& rhs )
{
   bool fv;

   if ( lhs.getYear() < rhs.getYear() )           fv = true;
   else if ( lhs.getYear() > rhs.getYear() )      fv = false;
   else // lhs.getYear() == rhs.getYear()
   {
      if ( lhs.getMonth() < rhs.getMonth() )      fv = true;
      else if ( lhs.getMonth() > rhs.getMonth() ) fv = false;
      else // lhs.getYear() == rhs.getMonth() &&
           // lhs.getMonth() == rhs.getMonth()
         fv = ( lhs.getDay() < rhs.getDay() );
   }
   return fv;
} // operator<( const Date

bool operator>=( const Date& lhs, const Date& rhs )
{  return !(lhs < rhs);  }

/*
bool operator==( const Date& lhs, const Date& rhs )
{
   return ( lhs.getYear() == rhs.getYear() &&
            lhs.getMonth() == rhs.getMonth() &&
            lhs.getDay() == rhs.getDay() );
}; // operator==( const Date

bool operator==( const Date& lhs, const Date& rhs )
{
   return ( lhs.dYear == rhs.dYear && lhs.dMonth == rhs.dMonth &&
            lhs.dDay == rhs.dDay );
}; // operator==( const Date
*/
#endif // _DATE_Cpp
